#pragma once

#define VULKAN_HPP_NO_CONSTRUCTORS
#include <vulkan/vulkan.hpp>

#include "../structs/queue_family_indices.hpp"
#include "../util/types.hpp"

class DeviceManager {
 public:
  DeviceManager() = default;

  /**
   * @brief Picks a suitable physical device (GPU) for the application.
   *
   * @param instance Vulkan instance to use for device selection
   * @param surface Surface to check for compatibility
   * @param requiredExtensions List of required device extensions
   * @return std::tuple<vk::PhysicalDevice, vk::SampleCountFlagBits, f32, bool> Physical device, MSAA samples,
   * max line width, wide line support
   */
  static fn pickPhysicalDevice(
    const vk::Instance&             instance,
    const vk::SurfaceKHR&           surface,
    const std::vector<const char*>& requiredExtensions
  ) -> std::tuple<vk::PhysicalDevice, vk::SampleCountFlagBits, f32, bool>;

  /**
   * @brief Creates a logical device with the specified features.
   *
   * @param physicalDevice Physical device to create logical device from
   * @param surface Surface to create device with
   * @param requiredExtensions Device extensions to enable
   * @param enableValidationLayers Whether to enable validation layers
   * @param validationLayers Validation layers to enable if enabled
   * @return std::tuple<vk::UniqueDevice, vk::Queue, vk::Queue> Logical device and queues
   */
  static fn createLogicalDevice(
    const vk::PhysicalDevice&       physicalDevice,
    const vk::SurfaceKHR&           surface,
    const std::vector<const char*>& requiredExtensions,
    bool                            enableValidationLayers,
    const std::vector<const char*>& validationLayers
  ) -> std::tuple<vk::UniqueDevice, vk::Queue, vk::Queue>;

  /**
   * @brief Gets the maximum supported MSAA sample count.
   *
   * @param physicalDevice Physical device to check
   * @return vk::SampleCountFlagBits Maximum supported sample count
   */
  static fn getMaxUsableSampleCount(const vk::PhysicalDevice& physicalDevice) -> vk::SampleCountFlagBits;

  /**
   * @brief Finds queue families that support required operations.
   *
   * @param device Physical device to check
   * @param surface Surface to check presentation support
   * @return QueueFamilyIndices Indices of suitable queue families
   */
  static fn findQueueFamilies(const vk::PhysicalDevice& device, const vk::SurfaceKHR& surface)
    -> QueueFamilyIndices;

  /**
   * @brief Checks if a physical device is suitable for the application.
   *
   * @param device Physical device to check
   * @param surface Surface to check compatibility with
   * @param requiredExtensions Required device extensions
   * @return bool True if device is suitable
   */
  static fn isDeviceSuitable(
    const vk::PhysicalDevice&       device,
    const vk::SurfaceKHR&           surface,
    const std::vector<const char*>& requiredExtensions
  ) -> bool;

  /**
   * @brief Checks if a physical device supports required extensions.
   *
   * @param device Physical device to check
   * @param requiredExtensions Required device extensions
   * @return bool True if device supports all required extensions
   */
  static fn checkDeviceExtensionSupport(
    const vk::PhysicalDevice&       device,
    const std::vector<const char*>& requiredExtensions
  ) -> bool;
};
