#include <vkfw.hpp>

#include "window_manager.hpp"

#include "../util/constants.hpp"

using namespace constants;

fn WindowManager::create(const char* title, const WindowCallbacks& callbacks)
  -> std::tuple<vkfw::UniqueInstance, vkfw::UniqueWindow> {
  // Initialize GLFW
  vkfw::UniqueInstance instance = vkfw::initUnique();

  // Don't create an OpenGL context
  vkfw::WindowHints hints;
  hints.clientAPI = vkfw::ClientAPI::eNone;

#ifdef __APPLE__
  // Required for macOS
  hints.cocoaRetinaFramebuffer = true;
#endif

  // Create window
  vkfw::UniqueWindow window = vkfw::createWindowUnique(WIDTH, HEIGHT, title, hints);

  // Get the primary monitor and its resolution
  vkfw::Monitor      primaryMonitor = vkfw::getPrimaryMonitor();
  const GLFWvidmode* videoMode      = primaryMonitor.getVideoMode();

  // Calculate window position to center it
  i32 xpos = (videoMode->width - WIDTH) / 2;
  i32 ypos = (videoMode->height - HEIGHT) / 2;

  // Set window position
  window->setPos(xpos, ypos);

  // Configure cursor for FPS-style camera control
  window->set<vkfw::InputMode::eCursor>(vkfw::CursorMode::eDisabled);

  // Set up callbacks
  if (callbacks.on_cursor_move)
    window->callbacks()->on_cursor_move = callbacks.on_cursor_move;

  if (callbacks.on_key)
    window->callbacks()->on_key = callbacks.on_key;

  if (callbacks.on_mouse_button)
    window->callbacks()->on_mouse_button = callbacks.on_mouse_button;

  if (callbacks.on_window_resize)
    window->callbacks()->on_window_resize = callbacks.on_window_resize;

  return std::make_tuple(std::move(instance), std::move(window));
}

fn WindowManager::getRequiredExtensions() -> std::vector<const char*> {
  // Get the required extensions from GLFW
  std::span<const char*> extensionsSpan = vkfw::getRequiredInstanceExtensions();
  return { extensionsSpan.begin(), extensionsSpan.end() };
}

fn WindowManager::getFramebufferSize(const vkfw::Window& window) -> std::pair<i32, i32> {
  return window.getFramebufferSize();
}

fn WindowManager::shouldClose(const vkfw::Window& window) -> bool { return window.shouldClose(); }

fn WindowManager::setTitle(const vkfw::Window& window, const std::string& title) -> void {
  window.setTitle(title);
}

fn WindowManager::setCursorMode(const vkfw::Window& window, vkfw::CursorMode mode) -> void {
  window.set<vkfw::InputMode::eCursor>(mode);
}

fn WindowManager::setPosition(const vkfw::Window& window, i32 xpos, i32 ypos) -> void {
  window.setPos(xpos, ypos);
}
